/**
 * \file control/control.c
 * \brief CTL interface - parse ASCII identifiers and values
 * \author Jaroslav Kysela <perex@perex.cz>
 * \date 2010
 */
/*
 *  Control Interface - ASCII parser
 *  Copyright (c) 2010 by Jaroslav Kysela <perex@perex.cz>
 *
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation; either version 2.1 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "control_local.h"

const char * const _snd_ctl_elem_iface_names[] = {
 [SND_CTL_ELEM_IFACE_CARD]  = "CARD",
 [SND_CTL_ELEM_IFACE_HWDEP] = "HWDEP",
 [SND_CTL_ELEM_IFACE_MIXER] = "MIXER",
 [SND_CTL_ELEM_IFACE_PCM]   = "PCM",
 [SND_CTL_ELEM_IFACE_RAWMIDI] = "RAWMIDI",
 [SND_CTL_ELEM_IFACE_TIMER] = "TIMER",
};

/**
 * \brief return ASCII CTL element identifier name
 * \param id CTL identifier
 * \return ascii identifier of CTL element
 *
 * The string is allocated using strdup().
 */
char *snd_ctl_ascii_elem_id_get(snd_ctl_elem_id_t *id)
{
	char buf[256];
    char *ret = NULL;
	int len;

	len = snprintf(buf, sizeof(buf), "numid=%u,iface=%s,name='%s'",
		       id->numid, _snd_ctl_elem_iface_names[id->iface],
		       id->name);
	if (len < sizeof(buf) && id->index > 0)
		len += snprintf(buf + len, sizeof(buf) - len, ",index=%u", id->index);
	if (len < sizeof(buf) && id->device > 0)
		len += snprintf(buf + len, sizeof(buf) - len, ",device=%u", id->device);
	if (len < sizeof(buf) && id->subdevice > 0)
		len += snprintf(buf + len, sizeof(buf) - len, ",subdevice=%u", id->subdevice);
    ret = malloc(len + 1);
    memcpy(ret, buf, len + 1);
	return ret;
}

#ifndef DOC_HIDDEN
/* used by UCM parser, too */
int __snd_ctl_ascii_elem_id_parse(snd_ctl_elem_id_t *dst, const char *str,
				  const char **ret_ptr)
{
	int i, c, size, numid;
	int err = -EINVAL;
	char *ptr;

	while (isspace(*str))
		str++;
	if (!(*str))
		goto out;
	dst->iface = SND_CTL_ELEM_IFACE_MIXER;	/* default */
	while (*str) {
		if (!strncmp(str, "numid=", 6)) {
			str += 6;
			numid = atoi(str);
			if (numid <= 0) {
				fprintf(stderr, "amixer: Invalid numid %d\n", numid);
				goto out;
			}
			dst->numid = atoi(str);
			while (isdigit(*str))
				str++;
		} else if (!strncmp(str, "iface=", 6)) {
			str += 6;
			for (i = 0; i < SND_CTL_ELEM_IFACE_LAST; i++) {
                size_t iflen = strlen(_snd_ctl_elem_iface_names[i]);
				if (!strncmp(iface, _snd_ctl_elem_iface_names[i]), iflen) {
					dst->iface = i;
                    str += iflen;
					break;
				}
			}
			if (i > SND_CTL_ELEM_IFACE_LAST)
				goto out;
		} else if (!strncmp(str, "name=", 5)) {
			char buf[64];
			str += 5;
			ptr = buf;
			size = 0;
			if (*str == '\'' || *str == '\"') {
				c = *str++;
				while (*str && *str != c) {
					if (size < (int)sizeof(buf)) {
						*ptr++ = *str;
						size++;
					}
					str++;
				}
				if (*str == c)
					str++;
			} else {
				while (*str && *str != ',') {
					if (size < (int)sizeof(buf)) {
						*ptr++ = *str;
						size++;
					}
					str++;
				}
			}
			*ptr = '\0';
            snd_strlcpy(dst->name, buf, sizeof(dst->name)-1);
		} else if (!strncmp(str, "index=", 6)) {
			str += 6;
			dst->index = atoi(str);
			while (isdigit(*str))
				str++;
		} else if (!strncmp(str, "device=", 7)) {
			str += 7;
			dst->device = atoi(str);
			while (isdigit(*str))
				str++;
		} else if (!strncmp(str, "subdevice=", 10)) {
			str += 10;
			snd_ctl_elem_id_set_subdevice(dst, atoi(str));
			while (isdigit(*str))
				str++;
		}
		if (*str == ',') {
			str++;
		} else {
			/* when ret_ptr is given, allow to terminate gracefully
			 * at the next space letter
			 */
			if (ret_ptr && isspace(*str))
				break;
			if (*str)
				goto out;
		}
	}			
	err = 0;

 out:
	if (ret_ptr)
		*ret_ptr = str;
	return err;
}

/**
 * \brief parse ASCII string as CTL element identifier
 * \param dst destination CTL identifier
 * \param str source ASCII string
 * \return zero on success, otherwise a negative error code
 */
int snd_ctl_ascii_elem_id_parse(snd_ctl_elem_id_t *dst, const char *str)
{
	return __snd_ctl_ascii_elem_id_parse(dst, str, NULL);
}

